/*======================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>

#define fsQ_PIC_VALS 4

#include "fsQuick.h"

#define private static

typedef unsigned char byte;
typedef unsigned short ushort;
typedef unsigned int uint;
typedef unsigned long ulong;

typedef struct
{
    short flag;
    short xpix;
    short ypix;
    short llen;
    short zero;
}
QHEAD, *PQHEAD;

typedef struct
{
    int fd;
    char *op;
    QHEAD qh;
    short comp;
    short qlar;
}
QLS, *PQLS;

static char *CheckArg (char ***pav, int *pac)
{
    char *p;
    char **av;

    av = *pav;

    if (*++(*av))
    {
	p = *av;
    }
    else if (*(av + 1) && **(av + 1) != '-')
    {
	p = *++av;
	*pac = *pac - 1;
    }
    else
    {
	p = NULL;
    }
    *pav = av;
    return p;
}
/*
 * This is not the same as the chunky pixels in 2gif & 2tif; these are
 * _very_ chunky pixels (i.e. 1 byte per pixel for 256 colours, which is
 * a bit of an overkill here).
 */

static void MakeChunkyPix (byte * cpix, byte * qpix, short xpix)
{
    byte *q, *p;
    ushort k;
    ushort j;
    ushort l;
    ushort gb, rb, n;
    int llen, ns, m;

    llen = ((xpix + 3) & ~3);

    p = cpix;
    q = qpix;

    memset (cpix, '\0', llen);

    k = 7;
    l = 0;

    for (m = 0; m < xpix; m++)
    {
	j = 1 << k;
	gb = (*q & j);
	if (gb)
	{
	    *p |= 2;
	}

	rb = *(q + 1) & j;
	if (rb)
	{
	    *p |= 1;
	}

	p++;

	if (k == 0)
	{
	    k = 7;
	    q += 2;
	}
	else
	{
	    k--;
	}
    }
}

static byte gmap[256] =
{0, 120, 200, 255};

int DoFS (PQLS pq)
{
    byte *qldat;
    byte *pic, *p, *xdata = NULL;
    short k, y, j;
    short height, width;
    int llen;
    fsQTablesT *fsQTablesP;
    int isiz;
    FILE *fp;

    height = pq->qh.ypix;
    width = (pq->qh.xpix + 3) & ~3;

    if (pq->qlar)
    {
	height = height * pq->qlar / (pq->qlar - 1);
    }

    llen = width;

    if ((qldat = (byte *) malloc (pq->qh.llen)))
    {
	if ((pic = malloc ((isiz = llen * height))))
	{
	    isiz = isiz >> 3;
	    if ((xdata = malloc (isiz)))
	    {
		memset (xdata, '\0', isiz);
		for (p = pic, k = y = 0; k < height; k++)
		{
		    if (pq->qlar == 0 || (((k + 1) % pq->qlar) && y < pq->qh.ypix))
		    {
			read (pq->fd, qldat, pq->qh.llen);
			y++;
		    }
		    MakeChunkyPix (p, qldat, pq->qh.xpix);
		    p += llen;
		}

		fprintf (stderr, "Dithering...\n");

		fsQTablesP = fsQInit (1, 0);

		fsQuick (pic, gmap, (int) height, (int) width,
			 (fsQPicT *) xdata, fsQTablesP, (width >> 3));

		if (fp = fopen (pq->op, "w"))
		{
		    fputs ("P4\n", fp);
		    fputs ("# Image by fspic (c) Jonathan Hudson\n", fp);
		    fprintf (fp, "%d %d\n", width, height);
		    fwrite (xdata, isiz, 1, fp);
		    fclose (fp);
		}
		free ((char *) xdata);
	    }
	    free ((char *) pic);
	}
	free ((char *) qldat);
    }
    if (xdata == NULL)
    {
	fputs ("No memory to dither...\n", stderr);
    }
    return 0;
}

main (int ac, char **av)
{
    QLS qs;
    struct stat s;
    char c, *p;
    short n;

    char *ip = NULL;

    qs.op = NULL;
    qs.qlar = 4;

    while (++av, --ac)
    {
	if (**av == '-')
	{
	    switch (c = *++(*av))
	    {
#ifdef QDOS
		case 'q':
		    {
			extern char *_endmsg;

			_endmsg = NULL;
		    }
		    break;
#endif
		case 'r':
		case 'g':
		    if ((p = CheckArg (&av, &ac)) != NULL)
		    {
			n = atoi (p);
			if (n && n < 255)
			{
			    gmap[(c == 'r') ? 1 : 2] = n;
			}
		    }
		    break;

		case 'i':
		    qs.qlar = 0;
		    break;
		case 'a':
		    qs.qlar = 3;
		    break;
	    }
	}
	else
	{
	    if (ip == NULL)
	    {
		ip = *av;
	    }
	    else if (qs.op == NULL)
	    {
		qs.op = *av;
	    }
	}
    }

    if ((qs.fd = open (ip, O_RDONLY, 0)) > 0)
    {
	read (qs.fd, &qs.qh, sizeof (QHEAD));
	if (qs.qh.flag != 0x4afc)
	{
	    fstat (qs.fd, &s);
	    qs.qh.xpix = 512;
	    qs.qh.llen = 128;
	    qs.qh.ypix = (s.st_size + 127) / 128;
	    lseek (qs.fd, 0, 0);
	}

	fprintf (stderr, "FS'ing %s to %s (%d x %d) -> ", ip, qs.op,
		 qs.qh.xpix, qs.qh.ypix);
	fflush (stderr);
	DoFS (&qs);
	close (qs.fd);
    }
    else
    {
	perror ("input");
    }
}
