#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#ifdef _DCC
# include <sys/types.h>
#else
# ifdef __linux__
# define stricmp strcasecmp
#else
  typedef unsigned char u_char;
  typedef unsigned short u_short;
  typedef unsigned long u_long;
  typedef unsigned int u_int;
# endif
#endif

#define DEF_FNTX    16
#define DEF_FNTY    28
#define DEF_FONTSIZE 14336	/* (FNTX/8)*FNTY*256 in  */
#define DEF_LINESPACE_T  4
#define DEF_LINESPACE_B  2
#define DEF_CHARSPACE_R  1
typedef struct
{
    char flag[8];
    u_char xpix;
    u_char ypix;
    u_char first;
    u_char last;
    u_char top_l;
    u_char bot_l;
    u_char right;
    u_char mode;
    long  fsize;
    u_char font[4];
} QFONT;


static u_char mask0[] = {0x80,0x40,0x20,0x10,8,4,2,1};
static u_char mask1[] = {0xc0, 0x30, 0xc, 0x3};

#if defined(__linux__) || defined (__VMS)
u_short SH (u_short val)
{
    return (u_short) (val << 8) + (val >> 8);
}
ulong LG (ulong val)
{
    return (ulong) (((ulong) SH (val & 0xFFFF) << 16) |
		    (ulong) SH(val >> 16));
}
#endif

void MemErr (void)
{
    fputs("Out of memory !", stderr);
    exit(0);
}

QFONT *readfont (char *fondat)
{
    FILE *fp;
    QFONT *qfont;
    
    if ((fp = fopen (fondat, "r")))
    {
	QFONT q;
	size_t mem;

	fread (&q, sizeof (q) - 2, 1, fp);
	fseek (fp, 0, 0);
	if (memcmp (q.flag, "QfaxFont", 8) == 0)
	{
#if defined (__linux__) || defined (__VMS)
	    q.fsize = LG(q.fsize);
#endif
	    mem = sizeof (q) + q.fsize - 4;
	    qfont = malloc (mem);

	    if(qfont)
	    {
		fread (qfont, mem, 1, fp);
#if defined (__linux__) || defined (__VMS)
		qfont->fsize = LG(qfont->fsize);
#endif

	    }
	    else
	    {
		MemErr ();
	    }
	}
	else
	{
	    mem = sizeof (QFONT) - 4 + DEF_FONTSIZE;
	    if ((qfont = malloc (mem)))
	    {
		fread (qfont->font, DEF_FONTSIZE, 1, fp);
		qfont->xpix = DEF_FNTX;
		qfont->ypix = DEF_FNTY;
		qfont->top_l = DEF_LINESPACE_T;
		qfont->bot_l = DEF_LINESPACE_B;
		qfont->right = DEF_CHARSPACE_R;
		memcpy (qfont->flag, "QfaxFont", 8);
		qfont->first = 0;
		qfont->last = 255;
		qfont->fsize = DEF_FONTSIZE;
		qfont->mode = 0;
	    }
	    else
	    {
		MemErr ();
	    }
	}
	fclose (fp);

	if(qfont->mode || qfont->first || qfont->last != 255)
	{
	    fputs ("Only for monospaced 255 char fonts\n", stderr);
	    exit(0);
	}
	
	return qfont;
    }

    fprintf(stderr, "Font \"%s\" not found\n", fondat);
    exit(0);
}

static void DumpChar (short ch, QFONT *qf)
{
    short i, j, b;
    u_char *c;
    int r;
    
    r = ch * qf->xpix * qf->ypix;

    for(i = 0; i < qf->ypix; i+= 2)
    {
	c = qf->font + (r >> 3);
	b = (r & 7) >> 1;

	for (j = 0; j < qf->xpix; j += 2)
	{
	    if(*c & mask1[b])
	    {
		fputc ('#', stdout);
	    }
	    else
	    {
		fputc (' ', stdout);
	    }
	    b++;
	    if(b == 4)
	    {
		b = 0;
		c++;
	    }
        }
	fputc('\n', stdout);
	r += (qf->xpix << 1);
    }
}

static short MakeChar(char *txt, QFONT *qf, short ch, short n)
{
    short i, j, done;
    char *p;
    u_char *c;
    int r;

    r = ch * qf->xpix * qf->ypix + (n-1)*qf->xpix;
    c = qf->font + (r >> 3);
    j = r & 7;
    
    for(i = done = 0, p = txt; i < (qf->xpix * 2) && *p && !done; i++, p++)
    {
        if(i & 1)
        {
            if(*p != '|')
            {
                done = 1;
            }
        }
        else
        {
	    if(j == 8)
	    {
		j = 0;
		c++;
	    }
            if(*p != ' ')
            {
                *c |= mask0[j];
	    }
	    else
	    {
		*c &= ~mask0[j];
	    }
	    j++;
        }
    }
    done |= (i != 32);
    return done;
}                    

static int NewChar(FILE *fp, QFONT *qf)
{
    char txt[128],  *rtxt;
    short done = 0;
    short ln, cpos;
    int n;
    short last;
    
    ln = 0;
    last = qf->ypix+4;

    while(!done && fgets(txt, sizeof(txt), fp))
    {
        if((rtxt = strchr(txt, '\n')))
        {
            *rtxt = 0;
        }
        ln++;
        if(ln == 1)
        {
	    cpos = 0;
            n = sscanf (txt, "%hi", &cpos);
        }
        else if (ln == 2)
	{
	   n = strncmp(txt, "   1 2 3 4", 10);
           done = n;
        }
	else if(ln == 3 || ln == last)
	{
	    n = strncmp(txt, "--------", 8);
            done = n;
            if(ln == last)
            {
		ln = 0;
	    }
	}
	else
	{
	    n = strtol(txt, &rtxt, 10);
            if(n == ln - 3 && *rtxt == '|')
            {
		done = MakeChar(rtxt+1, qf, cpos, n);
            }
            else
            {
                done = 1;
            }
        }
    }
    return done;
}

static void DoReplace (char *av)
{
    char s[128], *p;
    FILE *fp;
    QFONT *qf;
    
    if(av == NULL)
    {
        fputs ("Definition File : ", stdout);
        av = gets (s);
    }
    if (av)
    {
        if((fp = fopen(av, "r")))
        {
            if(fgets(s, sizeof(s), fp))
            {
                if((p = strchr(s, '\n')))
                {
                    *p = '\0';
                }
                if((qf = readfont(s)))
                {
                    int fd2,mem;
                    NewChar(fp, qf);
                    fd2 = open(s, O_CREAT|O_WRONLY|O_TRUNC);
                    mem = sizeof(QFONT) - 4 + qf->fsize;
#if defined (__linux__) || defined (__VMS)
		    qf->fsize = LG(qf->fsize);
#endif
                    write(fd2, qf, mem);
                    close(fd2);
                }
                else
                {
                    perror("font file");
                }
	    }
	    else
	    {
	        puts("file error");
	    }
            fclose(fp);
	}
	else
	{
	    perror("definition file");
	}
    }
}

static void MoveFont (FILE *fp, QFONT *qfont1, int fd2)
{
    QFONT *qfont2;
    short ch1, ch2;
    short i,j,n;
    {
        u_char *p1, *p2;
        char txt[132];
        long mem;       
        mem = sizeof(QFONT)-4 + qfont1->fsize;
        if(!(qfont2 = malloc(mem)))
        {
            MemErr();
        }
        memcpy(qfont2, qfont1, mem);

        /* Format is NewValue OldValue */
        
        for(n = 0; fgets(txt, sizeof(txt), fp); )
        {
            if(*txt != '#' && sscanf(txt, "%hi %hi", &ch2, &ch1) == 2)
            {
                int p1c,p1b,r1;
                int p2c,p2b,r2;
                printf("Move %d to %d\n", ch1, ch2);
                
        	r1 = ch1 * qfont1->xpix * qfont1->ypix;
        	r2 = ch2 * qfont2->xpix * qfont2->ypix;

                p1c = r1 >> 3;
                p1b = r1 & 7;
                p2c = r2 >> 3;
                p2b = r2 & 7;
                p1 = qfont1->font+p1c;                
                p2 = qfont2->font+p2c;                

        	for (i = 0; i < qfont1->ypix; i++)
                {
                    for(j = 0; j < qfont1->xpix; j++,p1b++,p2b++)
                    {
                        if(p1b == 8)
                        {
                            p1++;
                            p1b = 0;
                        }
                        if(p2b == 8)
                        {
                            p2++;
                            p2b = 0;
                        }
                        if(*p1 & mask0[p1b])
                        {
                            *p2 |= mask0[p2b];
                        }
                        else
                        {
                            *p2 &= ~mask0[p2b];
                        }
                    }
	        }
                n++;
	    }
	}
        if(n)
        {
#if defined (__linux__) || defined (__VMS)
    	    qfont2->fsize = LG(qfont2->fsize);
#endif
            write(fd2, qfont2, mem);
        }
    }
} 

static void DoChange (char *av)
{
    char s[128];
    FILE *fp;
    char *p;
    QFONT* qfont1;
    int fd2;
    
    if(av == NULL)
    {
        fputs ("Definition File : ", stdout);
        av = gets (s);
    }
    if (av)
    {
        if((fp = fopen(av, "r")))
        {
            if(fgets(s, sizeof(s), fp))
            {
                if((p = strchr(s, '\n')))
                {
                    *p = '\0';
                }
                qfont1 = readfont(s);
                if(qfont1 > 0)
                {
                    if(fgets(s, sizeof(s), fp))
                    {
                        if((p = strchr(s, '\n')))
                        {
                            *p = '\0';
                        }
                        fd2 = open(s, O_CREAT | O_EXCL | O_WRONLY);
                        if(fd2 > 0)
                        {
                            MoveFont (fp, qfont1, fd2);
                            close (fd2);
                            printf("New font file is %s\nDon't forget to rename before use !\n", s);
                        }
                        else
                        {
                            perror("New font");
                        }
                    }
                }
                else
                {
                    perror("Old font");
                }
            }
            else
            {
                puts("No in font");
            }
            fclose(fp);
        }
        else
        {
            perror("Def File");
        }
    }
}

static void DoView (char *av)
{
    int c;
    QFONT *qf;
    char s[128];

    if(av == NULL)
    {
        fputs ("Font File : ", stdout);
        av = gets (s);
    }
    if (av)
    {
	if ((qf = readfont(av)))
	{
	    while (fputs ("Value : ", stdout), gets (s))
	    {
		c = strtol (s, NULL, 0);
		if(c && c < 256)
		    DumpChar(c, qf);
	    }
	    fputc('\n', stdout);
	}
    }
}

int main (int ac, char **av)
{
    ac--; av++;
    if (ac && stricmp (*av, "view") == 0)
    {
#ifdef QDOS
        extern char *_endmsg;
        _endmsg = NULL;
#endif
	DoView (*(av+1));
    }
    else if (ac && stricmp (*av, "change") == 0)
    {
        DoChange(*(++av));
    }
    else if (ac && stricmp (*av, "replace") == 0)
    {
        DoReplace (*(++av));
    }
    else
    {
	puts ("VIEW or CHANGE or REPLACE");
    }
    return 0;
}
